@@ -31,7 +31,7 @@ open Ocamlbuild_plugin
 
 let try_exec command =
   try
-    let _ = run_and_read command in
+    let () = Command.execute ~quiet:true (Cmd(S[Sh command; Sh"> /dev/null"; Sh"2> /dev/null"])) in
     true
   with _ ->
     false
@@ -46,11 +46,14 @@ let have_native = try_exec "ocamlfind ocamlopt -version"
 let have_threads = try_exec "ocamlfind query threads"
 let have_ssl = try_exec "ocamlfind query ssl"
 let have_glib = try_exec "ocamlfind query lablgtk2" && try_exec "pkg-config glib-2.0"
-let have_text = try_exec "ocamlfind query text"
+let have_react = try_exec "ocamlfind query react"
+let have_text = have_react && try_exec "ocamlfind query text"
+
+let stdlib_path = String.chomp (run_and_read "ocamlfind ocamlc -where")
 
 (* Try to find the path where compiler libraries are: *)
 let compiler_libs =
-  let stdlib = String.chomp (run_and_read "ocamlfind ocamlc -where") in
+  let stdlib = String.chomp stdlib_path in
   try
     let path =
       List.find Pathname.exists [
@@ -75,6 +78,7 @@ let () =
 | preemptive threads support:   %3s |
 | ssl support:                  %3s |
 | glib support:                 %3s |
+| react support                 %3s |
 | text support:                 %3s |
 | super toplevel:               %3s |
 +-----------------------------------+
@@ -82,9 +86,71 @@ let () =
     (yes_no have_threads)
     (yes_no have_ssl)
     (yes_no have_glib)
+    (yes_no have_react)
     (yes_no have_text)
     (yes_no have_toplevel)
 
+type library = {
+  name : string;
+  (* The short name of the library ("core", "unix", ... *)
+  deps : string list;
+  (* Dependencies on other lwt libraries *)
+  have : bool;
+  (* Can we build it ? *)
+  test : bool;
+  (* Are they tests for it ? *)
+}
+
+(* All libraries with their dependencies *)
+let libraries = [
+  { name = "core";
+    have = true;
+    deps = [];
+    test = true };
+  { name = "unix";
+    have = true;
+    deps = ["core"];
+    test = true };
+  { name = "preemptive";
+    have = have_threads;
+    deps = ["core"; "unix"];
+    test = false };
+  { name = "extra";
+    have = have_threads;
+    deps = ["core"; "unix"; "preemptive"];
+    test = false };
+  { name = "glib";
+    have = have_glib;
+    deps = ["core"; "unix"];
+    test = false };
+  { name = "react";
+    have = have_react;
+    deps = ["core"];
+    test = true };
+  { name = "ssl";
+    have = have_ssl;
+    deps = ["core"; "unix"];
+    test = false };
+  { name = "text";
+    have = have_text;
+    deps = ["core"; "unix"; "react"];
+    test = false };
+  { name = "simple_top";
+    have = true;
+    deps = ["core"; "unix"];
+    test = false };
+  { name = "top";
+    have = have_text;
+    deps = ["core"; "unix"; "text"; "react"];
+    test = false };
+]
+
+let lib_path name suffix =
+  if name = "core" then
+    sprintf "src/core/lwt.%s" suffix
+  else
+    sprintf "src/%s/lwt_%s.%s" name name suffix
+
 (* +-----------------------------------------------------------------+
    | Ocamlfind                                                       |
    +-----------------------------------------------------------------+ *)
@@ -166,7 +232,7 @@ let pkg_config flags package =
 
 let define_stubs name =
   let tag = sprintf "use_%s_stubs" name in
-  dep ["link"; "ocaml"; tag] [sprintf "src/liblwt_%s_stubs.a" name];
+  dep ["link"; "ocaml"; tag] [sprintf "src/%s/stubs/liblwt_%s_stubs.a" name name];
   flag ["link"; "library"; "ocaml"; tag] & S[A"-cclib"; A(sprintf "-llwt_%s_stubs" name)];
   flag ["link"; "library"; "ocaml"; "byte"; tag] & S[A"-dllib"; A(sprintf "-llwt_%s_stubs" name)]
 
@@ -198,29 +264,38 @@ let _ =
         Options.ocamlopt := ocamlfind "ocamlopt";
         Options.ocamldep := ocamlfind "ocamldep";
         Options.ocamlmktop := ocamlfind "ocamlmktop";
+        (* possibility to override ocamldoc (for the website) *)
+        let ocamldoc =
+          try Sh (Sys.getenv "OCAMLDOC")
+          with Not_found -> S[A"ocamlfind"; A"ocamldoc"]
+        in
         (* FIXME: sometimes ocamldoc say that elements are not found
            even if they are present: *)
-        Options.ocamldoc := S[A"ocamlfind"; A"ocamldoc"; A"-hide-warnings"]
+        Options.ocamldoc := S[ocamldoc; A"-hide-warnings"]
 
     | After_rules ->
-        Pathname.define_context "src" [ "src/private" ];
-        Pathname.define_context "src/private" [ "src" ];
+        List.iter
+          (fun lib ->
+             let deps = List.map (sprintf "src/%s") lib.deps in
+             Pathname.define_context (sprintf "src/%s" lib.name) (sprintf "src/%s/private" lib.name :: deps);
+             Pathname.define_context (sprintf "src/%s/private" lib.name) (sprintf "src/%s" lib.name :: deps);
+             Pathname.define_context (sprintf "tests/%s" lib.name) ("tests" :: "src/core" :: "src/unix" :: sprintf "src/%s" lib.name :: sprintf "src/%s/private" lib.name :: deps))
+          libraries;
+
+        Pathname.define_context "tests" ["src/core"; "src/unix"];
 
         (* +---------------------------------------------------------+
            | Virtual targets                                         |
            +---------------------------------------------------------+ *)
 
-        let libs = "lwt" :: "lwt_unix" :: "simple_top" :: List.concat
-          (List.map snd
-             (List.filter fst
-                [(have_threads, ["lwt_preemptive"; "lwt_extra"]);
-                 (have_ssl, ["lwt_ssl"]);
-                 (have_glib, ["lwt_glib"]);
-                 (have_text, ["lwt_text"; "lwt_top"])])) in
+        let libs = List.filter_opt (fun lib -> if lib.have then Some lib.name else None) libraries in
 
-        let byte = "syntax/pa_lwt.cmo" :: List.map (sprintf "src/%s.cma") libs
-          @ if have_toplevel then ["src/toplevel.top"] else []
-        and native = List.map (sprintf "src/%s.cmxa") libs in
+        let byte =
+          "syntax/pa_lwt.cmo" :: "syntax/pa_log.cmo" ::
+            List.map (fun name -> lib_path name "cma") libs
+          @ if have_toplevel then ["src/top/private/toplevel.top"] else []
+        and native =
+          List.map (fun name -> lib_path name "cmxa") libs in
 
         let virtual_rule name deps =
           rule name ~stamp:name ~deps (fun _ _ -> Nop)
@@ -230,6 +305,15 @@ let _ =
         virtual_rule "byte" & "META" :: byte;
         virtual_rule "native" & "META" :: native;
 
+        let tests = List.filter_opt (fun lib -> if lib.have && lib.test then Some lib.name else None) libraries in
+
+        let tests = List.map (if have_native then
+                                sprintf "tests/%s/main.native"
+                              else
+                                sprintf "tests/%s/main.byte") tests in
+
+        virtual_rule "test_programs" & tests;
+
         (* +---------------------------------------------------------+
            | Internal syntaxes                                       |
            +---------------------------------------------------------+ *)
@@ -241,7 +325,8 @@ let _ =
 
              (* Make them depends on the syntax extension *)
              dep ["ocaml"; "ocamldep"; tag] [file])
-          [("pa_lwt", "syntax/pa_lwt.cmo")];
+          [("pa_lwt", "syntax/pa_lwt.cmo");
+           ("pa_log", "syntax/pa_log.cmo")];
 
         (* +---------------------------------------------------------+
            | Ocamlfind stuff                                         |
@@ -272,8 +357,47 @@ let _ =
         flag_all_stages "use_compiler_libs" & S(List.map (fun path -> S[A"-I"; A path]) compiler_libs);
 
         (* Link with the toplevel library *)
-        dep ["src/toplevel.top"] ["src/lwt.cma"; "src/lwt_unix.cma"; "src/lwt_text.cma"; "src/lwt_top.cma"];
-        flag ["file:src/toplevel.top"] & S[A"-I"; A"src"; A"lwt.cma"; A"lwt_unix.cma"; A"lwt_text.cma"; A"lwt_top.cma"];
+        let libs = ["core"; "react"; "unix"; "text"; "top"] in
+        dep ["file:src/top/private/toplevel_temp.top"] (List.map (fun name -> lib_path name "cma") libs);
+        flag ["file:src/top/private/toplevel_temp.top"] & S(A"-I" :: A"src/unix/stubs" :: List.map (fun name -> A(lib_path name "cma")) libs);
+
+        (* Expunge compiler modules *)
+        rule "toplevel expunge"
+          ~dep:"src/top/private/toplevel_temp.top"
+          ~prod:"src/top/private/toplevel.top"
+          (fun _ _ ->
+             let directories =
+               stdlib_path
+               :: "src/core"
+               :: "src/react"
+               :: "src/unix"
+               :: "src/text"
+               :: "src/top"
+               :: (List.map
+                    (fun lib ->
+                       String.chomp
+                         (run_and_read
+                            ("ocamlfind query " ^ lib)))
+                    ["findlib"; "react"; "unix"; "ssl"; "text"])
+             in
+             let modules =
+               List.fold_left
+                 (fun set directory ->
+                    List.fold_left
+                      (fun set fname ->
+                         if Pathname.check_extension fname "cmi" then
+                           StringSet.add (module_name_of_pathname fname) set
+                         else
+                           set)
+                      set
+                      (Array.to_list (Pathname.readdir directory)))
+                 StringSet.empty directories
+             in
+             Cmd(S[A(stdlib_path / "expunge");
+                   A"src/top/private/toplevel_temp.top";
+                   A"src/top/private/toplevel.top";
+                   A"outcometree"; A"topdirs"; A"toploop";
+                   S(List.map (fun x -> A x) (StringSet.elements modules))]));
 
         (* +---------------------------------------------------------+
            | C stubs                                                 |
@@ -286,15 +410,6 @@ let _ =
           define_c_library ~name:"glib" ~c_name:"glib-2.0"
         end;
 
-        dep ["use_config_h"] ["src/config.h"];
-        rule "config.h" ~prod:"src/config.h"
-          (fun _ _ ->
-             Echo((if try_exec "echo '#include <sys/signalfd.h>' | cpp" then
-                     ["#define HAVE_SIGNALFD 1\n"]
-                   else
-                     []),
-                  "src/config.h"));
-
         (* +---------------------------------------------------------+
            | Other                                                   |
            +---------------------------------------------------------+ *)
@@ -305,18 +420,41 @@ let _ =
              Echo([substitute [("@VERSION@", get_version ())] (read_file "META.in")], "META"));
 
         (* Generation of the lwt.odocl file *)
-        let deps = ["src/lwt.mllib"; "src/lwt_unix.mllib"; "src/lwt_preemptive.mllib"; "src/lwt_extra.mllib"; "src/lwt_ssl.mllib"]
+        let deps =
+          List.filter_opt
+            (fun lib ->
+               if lib.have then
+                 Some(lib_path lib.name "mllib")
+               else
+                 None)
+            libraries;
         and prod = "lwt.odocl" in
         rule "lwt_doc" ~prod ~deps
-          (fun _ _ -> Echo(List.map (sprintf "src/%s\n")
-                             (* Filter deprecated modules: *)
-                             (List.filter (function
-                                             | "Lwt_chan" -> false
-                                             | s -> not (String.is_prefix "private" s))
-                                (List.concat (List.map string_list_of_file deps)))
-                           @ ["syntax/Pa_lwt\n"],
+          (fun _ _ -> Echo(List.concat
+                             (List.map
+                                (fun lib ->
+                                   if lib.have then
+                                     List.map (sprintf "src/%s/%s\n" lib.name)
+                                       (List.filter (function
+                                                       | "Lwt_chan" | "Lwt_util" -> false
+                                                       | s -> not (String.is_prefix "private" s))
+                                          (string_list_of_file (lib_path lib.name "mllib")))
+                                   else
+                                     [])
+                                libraries)
+                           @ ["syntax/Pa_lwt\n"; "syntax/Pa_log"],
                            prod));
 
+        (* Use an introduction page with categories *)
+        tag_file "lwt.docdir/index.html" ["apiref"];
+        dep ["apiref"] ["apiref-intro"];
+        flag ["apiref"] & S[A "-intro"; P "apiref-intro"; A"-colorize-code"];
+
+        (* Build documentation then copy our css to the documentation
+           directory *)
+        rule "Documentation with custom css" ~deps:["utils/style.css"; "lwt.docdir/html.stamp"] ~stamp:"doc"
+          (fun _ _ -> cp "utils/style.css" "lwt.docdir/style.css");
+
         (* The default "thread" tag is not compatible with ocamlfind.
            Indeed, the default rules add the "threads.cma" or
            "threads.cmxa" options when using this tag. When using the
